#Need to expand for all regions, parallelize via foreach
#Ideally, this would work entirely from the daily data and
#aggregate to weekly as necessary


library("ggplot2")
library("foreach")
library("doParallel")
library("remotes")
library("forecast")
#library("forecastW2T")
#source('D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Kai_functions.R')
devtools::load_all('D:/OneDrive/Research/where2test/forecastW2T')

#Register a cluster for parallelized loop
nCores <- detectCores()
cl<- makeCluster(nCores[1]-1) #not to overload your computer
registerDoParallel(cl)

setwd("D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/")

#Weekly data
WDATA <- read.csv(file="Data/cases[RKI]_saxony_weekly.csv")

#Daily data
DDATA <- read.csv(file="Data/cases[RKI]_saxony_daily.csv")

CODES <- unique(DDATA$region)
L.CODES <- length(CODES)

#Length of the forecast horizon in weeks
HW <- 4

#Get the weekday of the latest observation and then the prediction periods
#Sunday is 0, Monday to Saturday is 1-6
exclude_days <- 3
last_weekday <- as.POSIXlt(DDATA$date)$wday[nrow(DDATA)-exclude_days]

HD <- HW*7 - last_weekday

#Compute the mixture parameters for daily forecasts
MODS_daily <- mix_daily(DDATA, H=HD, CORES=nCores[1]-1)

#Loop over regions
foreach(i=1:L.CODES) %dopar% {
  library("forecast")
  devtools::load_all('D:/OneDrive/Research/where2test/forecastW2T')
  #source('D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Kai_functions.R')

  #Get daily data for local region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  #Exclude last few days
  DDATASET <- head(DDATASET, n = -exclude_days)

  #Extract dates and case numbers
  DDATES <- as.Date(DDATASET$date)
  DCASES <- DDATASET$cases

  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)

  #Fit the SARIMA model with an automatic box-cox transformation
  SIM <- combination_daily(tail(DCASES, n=100), h=HD, m=MODS_daily[i, 2])

  #Extract the observations of the final week in the data
  PREWEEK <- tail(DCASES, n=last_weekday)
  
  #Create the nowcast by combining the observed data and the predictions
  NOWCAST <- c(PREWEEK, SIM$mean)

  #Get the mean for the nowcast
  NCM <- NOWCAST

  #Prediction means and lower and upper interval limits
  PDM.daily <- NCM
  PDL.daily <- c(rep(0,last_weekday),SIM$lower) 
  PDU.daily <- c(rep(0,last_weekday),SIM$upper) 

  #Daily data stuff starts here
  #Get the end date of the data
  ED <- DDATES[length(DDATES)]

  #Generate a vector of forecast dates
  DATES.PRD.daily <- seq(ED+1, ED+HD, 1)
  
  #Create output dataframes for data and forecasts
  L.DDATA <- length(DCASES)

  #Fill in prediction interval columns with NAs to keep same structure as forecast output
  DATA.OUT <- data.frame(
    region = rep(CODES[i], L.DDATA),
    date = DDATES,
    m = DDATASET$cases,
    l95 = rep(NA, L.DDATA),
    u95 = rep(NA, L.DDATA)
  )

  FORE.OUT <- data.frame(
    region = rep(CODES[i], HD),
    date = DATES.PRD.daily,
    m = SIM$mean,
    l95 = SIM$lower[,1],
    u95 = SIM$upper[,1]
  )

  #Collate the data and forecast outputs
  FORECASTS <- rbind(DATA.OUT, FORE.OUT)

  if (!dir.exists("Predictions")) {dir.create("Predictions")}
  if (!dir.exists("Figures")) {dir.create("Figures")}
  write.csv(FORECASTS, file=paste0("Predictions/Daily_", toString(CODES[i])), row.names = F)

  #Need to output figure via graphics device
  png(file=paste0("Figures/Daily_", toString(CODES[i]), ".png"),
      width=1200, height=700)

  par(mar=c(5,5,2,0))

  #Plot daily data
  plot(DDATES, DCASES, xaxt='n', type="l", ylab="Cases",
       xlab="Date", ylim=c(0, max(DCASES, PDU.daily)),
       xlim=as.Date(c(DDATES[1], tail(DATES.PRD.daily, n=1))),
       cex.lab=1.5, lwd=2, main=toString(CODES[i]))

  #Add the lines for the mean predictions
  lines(DATES.PRD.daily, SIM$mean, col="blue", lwd=2)

  #Add the PIs as a blue shaded region
  polygon(x = c(DATES.PRD.daily, rev(DATES.PRD.daily)),
          y = c(SIM$upper, rev(SIM$lower)),
          col =  adjustcolor("dodgerblue", alpha.f = 0.30), border = NA)

  #Add a red point to indicate the nowcast
  points(DATES.PRD.daily[1:7], NCM[1:7], col="red", lwd=2, pch=20)

  #Add the correct dates to the x-axis
  ALL.DATES <- seq(DDATES[1], DATES.PRD.daily[length(DATES.PRD.daily)],7)
  axis(1, ALL.DATES, format(ALL.DATES, "%m/%y"), cex.axis = 1)

  dev.off()

  #Aggregated to weekly predictions
  #Create the nowcast by combining the observed data and the predictions
  NOWCAST <- PDM.daily[1:7]

  #Get the mean and PI limits for the nowcast
  NCM <- sum(NOWCAST)
  NCL <- sum(PDL.daily[1:7]) #quantile(NOWCAST, 0.025)
  NCU <- sum(PDU.daily[1:7]) #quantile(NOWCAST, 0.975)

  #Aggregate the daily predictions for week 1 after the nowcast
  W1 <- sum(PDM.daily[8:14]) #rowSums(SIM[, 5:11])

  #Get the mean and PI limits for the week 1 forecast
  W1M <- W1 #mean(W1)
  W1L <- sum(PDL.daily[8:14]) #quantile(W1, 0.025)
  W1U <- sum(PDU.daily[8:14]) #quantile(W1, 0.975)

  #Aggregate the daily predictions for week 2 after the nowcast
  W2 <- sum(PDM.daily[15:21]) #rowSums(SIM[, 12:18])

  #Get the mean and PI limits for the week 2 forecast
  W2M <- W2 #mean(W2)
  W2L <- sum(PDL.daily[15:21]) #quantile(W2, 0.025)
  W2U <- sum(PDU.daily[15:21]) #quantile(W2, 0.975)

  #Aggregate the daily predictions for week 3 after the nowcast
  W3 <- sum(PDM.daily[22:28]) #rowSums(SIM[, 19:25])

  #Get the mean and PI limits for the week 3 forecast
  W3M <- W3 #mean(W3)
  W3L <- sum(PDL.daily[22:28]) #quantile(W3, 0.025)
  W3U <- sum(PDU.daily[22:28]) #quantile(W3, 0.975)

  #Prediction means and lower and upper interval limits
  PDM.weekly <- c(NCM, W1M, W2M, W3M)
  PDL.weekly <- c(NCL, W1L, W2L, W3L)
  PDU.weekly <- c(NCU, W1U, W2U, W3U)

  #Weekly data stuff starts here
  WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("date", "cases"))

  #THIS IS TEMPORARY. REMOVE ONCE DATA ARE IN CONSISTENT FORMAT
  WDATASET  <- WDATASET[-nrow(WDATASET), ]

  WCASES <- WDATASET$cases
  WDATES <- as.Date(WDATASET$date)

  #Get the end date of the data
  ED <- WDATES[length(WDATES)]

  #Generate a vector of forecast dates
  DATES.PRD.weekly <- c(ED + 7, ED + 14, ED + 21, ED + 28)

  #Create output dataframes for data and forecasts
  L.WDATA <- length(WCASES)

  #Fill in prediction interval columns with NAs to keep same structure as forecast output
  DATA.OUT <- data.frame(
    region = rep(CODES[i], L.WDATA),
    date = WDATES,
    m = WCASES,
    l95 = rep(NA, L.WDATA),
    u95 = rep(NA, L.WDATA)
  )

  FORE.OUT <- data.frame(
    region = rep(CODES[i], HW),
    date = DATES.PRD.weekly,
    m = PDM.weekly,
    l95 = PDL.weekly,
    u95 = PDU.weekly
  )

  #Collate the data and forecast outputs
  FORECASTS <- rbind(DATA.OUT, FORE.OUT)

  if (!dir.exists("Predictions")) {dir.create("Predictions")}
  if (!dir.exists("Figures")) {dir.create("Figures")}
  write.csv(FORECASTS, file=paste0("Predictions/Combine_", toString(CODES[i])), row.names = F)

  #Need to output figure via graphics device
  png(file=paste0("Figures/Combine_", toString(CODES[i]), ".png"),
      width=1200, height=700)

  par(mar=c(5,5,2,0))

  #Plot weekly data
  plot(c(WDATES,DATES.PRD.weekly[1]), c(WCASES,PDM.weekly[1]), xaxt='n', type="l", ylab="Cases",
       xlab="Date", ylim=c(0, max(WCASES, W3U)),
       xlim=as.Date(c(WDATES[1], tail(DATES.PRD.weekly, n=1))),
       cex.lab=1.5, lwd=2, main=toString(CODES[i]))

  #Add the lines for the mean predictions
  lines(DATES.PRD.weekly, PDM.weekly, col="blue", lwd=2)

  #Add the PIs as a blue shaded region
  polygon(x = c(DATES.PRD.weekly, rev(DATES.PRD.weekly)),
          y = c(PDU.weekly, rev(PDL.weekly)),
          col =  adjustcolor("dodgerblue", alpha.f = 0.30), border = NA)

  #Add a red point to indicate the nowcast
  points(DATES.PRD.weekly[1], NCM, col="red", lwd=2, pch=20)
  points(WDATES, WCASES, lwd=2, pch=20)
  points(DATES.PRD.weekly[2:4], PDM.weekly[2:4], lwd=2, pch=20)

  #Add the correct dates to the x-axis
  ALL.DATES <- c(WDATES, DATES.PRD.weekly)
  axis(1, ALL.DATES, format(ALL.DATES, "%B"), cex.axis = 1)

  dev.off()
}
stopCluster(cl)