farima <- function(x, h){
  forecast(Arima(x+1, order=c(0, 1, 2), seasonal=c(0,1,1), lambda="auto"), h, level=c(95))
}

fholt_a <- function(x, h){
  hw(x+1, h, damped=T, seasonal="additive", level=c(95), lambda="auto")
}

fholt_m <- function(x, h){
  hw(x+1, h, damped=T, seasonal="multiplicative", level=c(95), lambda="auto")
}

fsarima <- function(x, h){
  forecast(Arima(tail(x+1,n=100), order=c(0, 1, 2), seasonal=c(0,1,1), lambda="auto"), h, level=c(95))
}

fhw <- function(x, h){
  hw(tail(x+1,n=100), h, damped=T, seasonal="additive", level=c(95), lambda="auto")
}

mix_daily <- function(X, H, BCT=T, CORES, ini){

  #Register cores to use for parallelization
  #registerDoParallel(CORES)

  #Check if the data are on cases or tests. If on tests, note that and then change
  #the column title to cases, which is what the following code assumes
  if(names(X)[3]=="tests"){
    TESTS <- TRUE
    names(X)[3] <- "cases"
  } else{
    TESTS <- FALSE
  }

  #Extract the region codes from the data
  CODES <- unique(X$region)

  MODS <- foreach(i=CODES, .combine='rbind') %dopar% {
    #source('D:/OneDrive/Research/where2test/forecastw2t/R/errorplots.R')
    print(i)
    #Extract data for focal region
    DATASET <- subset(X, X$region == i, select=c("date", "cases"))

    #Extract dates and case numbers
    #DATES <- as.Date(DATASET$date)
    CASES <- DATASET$cases

    DDATES <- as.Date(DATASET$date)
    DCASES <- DATASET$cases
    DCASES <- ts(DCASES, start=c(as.numeric(format(DDATES[1], "%Y")), as.numeric(format(DDATES[1], "%j"))),
                 frequency = 7)

    #Create an empty vector to store optimization results
    RES <- c()

    #Grid search over values of the mixing parameter, m, in the combination forecast model
    for(j in seq(0, 1, 0.05)){

      #Calculate the raw prediction errors by rolling cross-validation
      #CV starts at week 28, which is at the beginning of the second wave
      CVE <- tsCV(DCASES, combination_daily, h=H, m=j, tests=TESTS, bct=BCT, initial=ini)

      #Error metric to minimize
      #Mean absolute prediction error, MAPE
      ERR <- mape(CVE, CASES)

      #Store model parameters and resulting error value
      RES <- rbind(RES, c(j, sum(ERR)))

    }

    #Identify and store the parameters associated with the min error
    MINROW <- which.min(RES[,2])
    MINM <- RES[MINROW, 1]

    #Create data structure to return to MODS
    c(i, MINM)
  }

  return(MODS)
}

combination_daily <- function(x, h, m, tests=FALSE, bct = TRUE) {

  #Automatic BoxCox transformation
  if(bct == T){

    x <- x + 1
    lbd <- BoxCox.lambda(x)
    x <- BoxCox(x, lambda=lbd)

  }

  #Set the ARIMA model based on whether the data represent tests or cases
  if(tests==TRUE){
    ORDER=c(0, 1, 0)
  } else {
    ORDER=c(1, 1, 0)
  }

  #Get individual arima and holt forecasts
  farima <- forecast(Arima(tail(x,n=100), order=c(0, 1, 2), seasonal=c(0,1,1), lambda="auto"), h, level=c(95))
  fholt <- hw(tail(x,n=100), h, damped=T, seasonal="additive", level=c(95), lambda="auto")

  #Output needs to be a forecast object, so use fitted arima model to get the
  #object structure, and overwrite the relevant details with values from combination model
  comb <- farima

  #Calculate the point forecast as a weighted average of the individual point forecasts
  comb$mean <- m*farima$mean + (1-m)*fholt$mean

  #Prediction intervals by weighted average of error variances
  #First back calculate the error variance from the CI interval end points
  ar_var <- ((farima$upper - farima$mean)/1.96)^2
  ht_var <- ((fholt$upper - fholt$mean)/1.96)^2

  #Calculate the combined variance as a weighted average of the components
  #Note the squaring of the weights
  cb_var <- m^2 * ar_var + (1-m)^2 * ht_var

  #Calculate the prediction interval endpoints under a Gaussian assumption
  #based on the combined error variance
  comb$lower <- comb$mean - 1.96 * sqrt(cb_var)
  comb$upper <- comb$mean + 1.96 * sqrt(cb_var)

  if(bct == T){

    #Back transform point and interval estimates and update the forecast object
    comb$mean <- InvBoxCox(comb$mean, lambda=lbd)
    comb$lower <- InvBoxCox(comb$lower, lambda=lbd)
    comb$upper <- InvBoxCox(comb$upper, lambda=lbd)

    #Back transform the data and update the forecast object
    x <- InvBoxCox(x, lambda=lbd)
    x <- x - 1
    comb$x <- x

  }

  #Give the model and appropriate name
  comb$method <- "Seasonal ARIMA-Holt Combination"

  #Return a forecast object for the predictions of the combination model
  return(comb)

}

