library("foreach")
library("doParallel")

#source('D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Kai_functions.R')
devtools::load_all('D:/OneDrive/Research/where2test/forecastW2T')

#define evaluation functions----
compute_daily_pred <- function(DDATA, WDATA, HD, pred_function, ini=0){
  CODES <- unique(WDATA$region)
  L.CODES <- length(CODES)
  
  DDATA$pred <- NA
  WDATA$pred <- NA
  
  Daily_pred_region <- foreach(i=1:L.CODES, .combine='rbind') %do% {
    #library("forecast")
    
    #i<-1
    #Get daily data for focal region
    DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("region", "date", "cases"))
    DDATASET$date <- as.Date(DDATASET$date)
    #DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
    
    #Extract dates and case numbers
    DDATA$pred <- NA
    WDATA$pred <- NA
    
    DDATASET <- DDATA
    DDATASET$date <- as.Date(DDATASET$date)
    
    #Extract dates and case numbers
    DDATES <- as.Date(DDATASET$date)
    DCASES <- DDATASET$cases
    
    #Cast the case counts as a timeseries with weekly frequency
    DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
                 frequency = 7)
    
    #Probably need to settle on a consistent arima model to use to avoid overfitting
    #Until then, use auto.arima
    #ARIMA models behave weirdly when they have too much data to chew on
    #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
    #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
    
    #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
    
    #This is the snaive model formulated as an arima model so that it is compatible with sim code below
    #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
    #summary(SARIMA)
    
    #Fit the SARIMA model with an automatic box-cox transformation
    #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
    #MODEL IS FOUND
    
    #Register a cluster for parallelized loop
    nCores <- detectCores()
    cl<- makeCluster(nCores[1]-1) #not to overload your computer
    registerDoParallel(cl)
    
    Daily_pred <- foreach(t=1:floor((length(DCASES)-100)/HD), .combine='rbind') %dopar% {
      devtools::load_all('D:/OneDrive/Research/where2test/forecastW2T')
      
      if (t>1) {
        DCASES_sub <- head(tail(DCASES, n = -28*(t-1)), n = 100)
        DDATASET_sub <- head(DDATASET, n = 100+28*(t-1))
      } else {
        DCASES_sub <- head(DCASES, n = 100)
        DDATASET_sub <- head(DDATASET, n = 100)
      }
      
      if (isTRUE(all.equal(pred_function, combination_daily, check.environment=FALSE))) {
        MODS_daily <- mix_daily(DDATASET_sub, H=HD, CORES=1, ini=ini)
        if (L.CODES == 1) MODS_daily <- t(data.frame(as.numeric(MODS_daily)))
        SIM <- pred_function(DCASES_sub, h=HD, m=MODS_daily[i, 2])$mean
      } else {
        SIM <- pred_function(DCASES_sub, h=HD)$mean
      }
      
      #Quick check if average (within time) prediction results look reasonable
      #plot(colMeans(SIM), type="l")
      
      NCM <- SIM #c(tail(DCASES_sub, n = (7-HDNC)), SIM)
      k <- 100 + (t-1)*28
      DDATA[DDATA$region == CODES[i],][(k+1):(k+28),'pred'] <- NCM
      WDATA[WDATA$region == CODES[i],][((t-1)*4+1):((t-1)*4+4),'pred'] <- colSums(matrix(NCM, nrow = 7))
      DDATA$period <- 'Daily'
      WDATA$period <- 'Weekly'
      rbind(DDATA[complete.cases(DDATA),], WDATA[complete.cases(WDATA),])
    }
    stopCluster(cl)
    Daily_pred
  }
  #list(DDATA[DDATA$region == CODES[i],], WDATA[WDATA$region == CODES[i],])
  return(Daily_pred_region)
}

compute_weekly_pred <- function(WDATA, HD, pred_function){
  #registerDoParallel(CORES)
  CODES <- unique(WDATA$region)
  L.CODES <- length(CODES)
  
  WDATA$pred <- NA
  
  Weekly_pred_region <- foreach(i=1:L.CODES, .combine='rbind') %do% {
    #library("forecast")
    
    #i<-1
    #Get daily data for focal region
    WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("region", "date", "cases"))
    WDATASET$date <- as.Date(WDATASET$date)
    
    #Extract dates and case numbers
    WDATES <- as.Date(WDATASET$date)
    WCASES <- WDATASET$cases
    
    #Cast the case counts as a timeseries with weekly frequency
    WCASES <- ts(WCASES, start=c(2020, as.numeric(format(WDATES[1], "%j"))),
                 frequency = 7)
    
    #Probably need to settle on a consistent arima model to use to avoid overfitting
    #Until then, use auto.arima
    #ARIMA models behave weirdly when they have too much data to chew on
    #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
    #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
    
    #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
    
    #This is the snaive model formulated as an arima model so that it is compatible with sim code below
    #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
    #summary(SARIMA)
    
    #Fit the SARIMA model with an automatic box-cox transformation
    #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
    #MODEL IS FOUND
    
    #Register a cluster for parallelized loop
    nCores <- detectCores()
    cl<- makeCluster(nCores[1]-1) #not to overload your computer
    registerDoParallel(cl)
    
    Weekly_pred <- foreach(t=1:floor((length(WCASES)-28-15)/HD), .combine='rbind') %dopar% {
      devtools::load_all('D:/OneDrive/Research/where2test/forecastW2T')
      
      WCASES_sub <- head(WCASES, n = 28+15+4*(t-1))
      WDATASET_sub <- head(WDATASET, n = 28+15+4*(t-1))
      
      if (isTRUE(all.equal(pred_function, combination, check.environment=FALSE))) {
        MODS <- mix(WDATASET_sub, H=HD)
        if (L.CODES == 1) MODS <- t(data.frame(as.numeric(MODS)))
        SIM <- pred_function(WCASES_sub, h=HD, m=MODS[i, 2])$mean
      } else {
        SIM <- pred_function(WCASES_sub, h=HD)$mean
      }
      
      NCM <- SIM #colMeans(SIM)
      k <- 4*(t-1)+28+15
      WDATA[WDATA$region == CODES[i],][(k+1):(k+4),'pred'] <- NCM
      WDATA[complete.cases(WDATA),]
    }
    stopCluster(cl)
    Weekly_pred
  }
  return(Weekly_pred_region)
}
#Register a cluster for parallelized loop
# nCores <- detectCores()
# cl<- makeCluster(nCores[1]-1) #not to overload your computer
# registerDoParallel(cl)

#main----
setwd("D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/")

eva_region <- 'saxony' #'dolnoslaski' #'bremen' #'ustecky' #
eva_region_id <- '14000' #'2' #'4' #'42' #

#Weekly data
WDATA_full <- read.csv(file=paste0("Data/cases[RKI]_", eva_region, "_weekly.csv"))
WDATA_full$date <- as.Date(WDATA_full$date)
WDATA_full <- WDATA_full[WDATA_full$region == eva_region_id,]
#WDATA <- WDATA[WDATA$date > "2020-12-20",]
#WDATA$pred <- NA

#Daily data
DDATA_full <- read.csv(file=paste0("Data/cases[RKI]_", eva_region, "_daily.csv"))
DDATA_full$date <- as.Date(DDATA_full$date)
DDATA_full <- DDATA_full[DDATA_full$region == eva_region_id,]

if (eva_region == 'dolnoslaski'){
  DDATA <- DDATA[-1,]
  DDATA$date <- as.Date(DDATA$date)
  fulldate <- data.frame(date=seq(DDATA$date[1], DDATA$date[nrow(DDATA)], by="days"))
  DDATA <- merge(DDATA,fulldate,by='date',all=T)
  DDATA$region <- '2'
  DDATA$population <- 2898525
  DDATA[is.na(DDATA$cases),'cases'] <- 0
}

# CODES <- unique(WDATA$region)
# L.CODES <- length(CODES)
# NOBS <- 100
# 
# #Number of sims. This should be large #(to reduce the simulation time, change 10000 to 100)
# NREP <- 10000
# 
# #Length of the forecast horizon in weeks, excluding nowcast
# HW <- 4
# 
# #Overall forecast horizon in days including nowcast
# HD <- 28
# 
# DDATA_forecast <- DDATA[DDATA$date > "2020-12-20",]
# DDATA_forecast$pred <- NA
# Daily_pred_sarima_10000 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
#   library("forecast")
#   
#   #i<-1
#   #Get daily data for focal region
#   DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
#   DDATASET$date <- as.Date(DDATASET$date)
#   DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
#   
#   #Extract dates and case numbers
#   DDATES <- as.Date(head(DDATASET$date, n=-6))
#   DCASES <- head(DDATASET$cases, n=-6)
#   
#   #Cast the case counts as a timeseries with weekly frequency
#   DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
#                frequency = 7)
#   
#   #Probably need to settle on a consistent arima model to use to avoid overfitting
#   #Until then, use auto.arima
#   #ARIMA models behave weirdly when they have too much data to chew on
#   #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
#   #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
#   
#   #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
#   
#   #This is the snaive model formulated as an arima model so that it is compatible with sim code below
#   #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
#   #summary(SARIMA)
#   
#   #Fit the SARIMA model with an automatic box-cox transformation
#   #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
#   #MODEL IS FOUND
#   
#   
#   t <- 0
#   n <- 1
#   DCASES_sub <- head(DCASES, n = 100)
#   while (length(DCASES_sub) == 100) {
#     SARIMA <- Arima(tail(DCASES_sub+1, n=100), order=c(0,1,2),
#                     seasonal=c(0,1,1), lambda="auto")
#     #Generate forecasts via simulation on a daily level with the SARIMA model
#     SIM <- NULL
#     for(j in 1:NREP) {
#       SIM <- rbind(SIM, simulate(SARIMA, future=TRUE, nsim=HD))
#     }
#     
#     #Quick check if average (within time) prediction results look reasonable
#     #plot(colMeans(SIM), type="l")
#     
#     #Extract the first 3 days of the final week in the data
#     PREWEEK <- matrix(tail(DCASES_sub, n=(7-HDNC)), nrow=NREP, ncol=7-HDNC, byrow=TRUE)
#     
#     #Create the nowcast by combining the last 4 days of data and the first 3 days of predictions
#     NOWCAST <- cbind(PREWEEK, SIM)
#     
#     NCM <- colMeans(NOWCAST)
#     DDATA_forecast[DDATA_forecast$region == CODES[i],][(t+1):(t+28),'pred'] <- NCM
#     DDATA_forecast[DDATA_forecast$region == CODES[i],][(t+1):(t+28),]
#     
#     DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
#     
#     t <- t + 28
#     n <- n + 1
#   }
#   DDATA_forecast[DDATA_forecast$region == CODES[i],]
# }
# 
# Weekly_pred_sarima_10000 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
#   library("forecast")
#   
#   #i<-1
#   #Get daily data for focal region
#   DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
#   DDATASET$date <- as.Date(DDATASET$date)
#   DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
#   
#   #Extract dates and case numbers
#   DDATES <- as.Date(head(DDATASET$date, n=-6))
#   DCASES <- head(DDATASET$cases, n=-6)
#   
#   #Cast the case counts as a timeseries with weekly frequency
#   DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
#                frequency = 7)
#   
#   #Probably need to settle on a consistent arima model to use to avoid overfitting
#   #Until then, use auto.arima
#   #ARIMA models behave weirdly when they have too much data to chew on
#   #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
#   #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
#   
#   #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
#   
#   #This is the snaive model formulated as an arima model so that it is compatible with sim code below
#   #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
#   #summary(SARIMA)
#   
#   #Fit the SARIMA model with an automatic box-cox transformation
#   #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
#   #MODEL IS FOUND
#   
#   
#   t <- 0
#   n <- 1
#   DCASES_sub <- head(DCASES, n = 100)
#   while (length(DCASES_sub) == 100) {
#     SARIMA <- Arima(tail(DCASES_sub+1, n=100), order=c(0,1,2),
#                     seasonal=c(0,1,1), lambda="auto")
#     #Generate forecasts via simulation on a daily level with the SARIMA model
#     SIM <- NULL
#     for(j in 1:NREP) {
#       SIM <- rbind(SIM, simulate(SARIMA, future=TRUE, nsim=HD))
#     }
#     
#     #Quick check if average (within time) prediction results look reasonable
#     #plot(colMeans(SIM), type="l")
#     
#     #Extract the first 3 days of the final week in the data
#     PREWEEK <- matrix(tail(DCASES_sub, n=(7-HDNC)), nrow=NREP, ncol=7-HDNC, byrow=TRUE)
#     
#     #Create the nowcast by combining the last 4 days of data and the first 3 days of predictions
#     NOWCAST <- cbind(PREWEEK, SIM)
#     
#     NCM <- colMeans(NOWCAST)
#     WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- colSums(matrix(NCM, nrow = 7))
#     WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
#     
#     DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
#     
#     t <- t + 4
#     n <- n + 1
#   }
#   WDATA[WDATA$region == CODES[i],]
# }

DDATA <- DDATA_full[DDATA_full$date > "2020-09-11",]
WDATA <- WDATA_full[WDATA_full$date > "2020-12-20",]
HD <- 28
Daily_pred_sarima <- compute_daily_pred(DDATA, WDATA, HD, fsarima)
Daily_pred_hw <- compute_daily_pred(DDATA, WDATA, HD, fhw)
Daily_pred_combine <- compute_daily_pred(DDATA, WDATA, HD, combination_daily, ini=0)
WDATA <- WDATA_full
HD <- 4
Weekly_pred_arima_weekly <- compute_weekly_pred(WDATA, HD, fsarima)
Weekly_pred_hw_weekly <- compute_weekly_pred(WDATA, HD, fhw)
Weekly_pred_comb_weekly <- compute_weekly_pred(WDATA, HD, combination)

Weekly_pred_sarima <- Daily_pred_sarima[Daily_pred_sarima$period=='Weekly',c('region','date','cases','pred')]
Weekly_pred_hw <- Daily_pred_hw[Daily_pred_hw$period=='Weekly',c('region','date','cases','pred')]
Weekly_pred_comb <- Daily_pred_combine[Daily_pred_combine$period=='Weekly',c('region','date','cases','pred')]

Daily_pred_hw_0 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  DDATASET$date <- as.Date(DDATASET$date)
  DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
  
  #Extract dates and case numbers
  DDATES <- as.Date(head(DDATASET$date, n=-6))
  DCASES <- head(DDATASET$cases, n=-6)
  
  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)
  
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  
  t <- 0
  n <- 1
  DCASES_sub <- head(DCASES, n = 100)
  while (length(DCASES_sub) == 100) {
    SIM <- hw(tail(DCASES_sub+1, n=100), h=HD, damped=T, seasonal="additive", lambda="auto")$mean
    
    NCM <- c(tail(DCASES_sub, n=(7-HDNC)), SIM)
    DDATA_forecast[DDATA_forecast$region == CODES[i],][(t+1):(t+28),'pred'] <- NCM
    DDATA_forecast[DDATA_forecast$region == CODES[i],][(t+1):(t+28),]
    
    DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
    
    t <- t + 28
    n <- n + 1
  }
  DDATA_forecast[DDATA_forecast$region == CODES[i],]
}

Weekly_pred_hw_0 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  DDATASET$date <- as.Date(DDATASET$date)
  DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
  
  #Extract dates and case numbers
  DDATES <- as.Date(head(DDATASET$date, n=-6))
  DCASES <- head(DDATASET$cases, n=-6)
  
  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)
  
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  
  t <- 0
  n <- 1
  DCASES_sub <- head(DCASES, n = 100)
  while (length(DCASES_sub) == 100) {
    SIM <- hw(tail(DCASES_sub+1, n=100), h=HD, damped=T, seasonal="additive", lambda="auto")$mean
    
    NCM <- c(tail(DCASES_sub, n=(7-HDNC)), SIM)
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- colSums(matrix(NCM, nrow = 7))
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
    
    DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
    
    t <- t + 4
    n <- n + 1
  }
  WDATA[WDATA$region == CODES[i],]
}

X <- DDATA[DDATA$date > as.Date('2020-09-14'),]
H <- 25
BCT <- TRUE
TESTS <- FALSE
#Extract the region codes from the data
CODES <- unique(X$region)

MODS_daily_0 <- foreach(i=CODES, .combine='rbind', .packages="forecast") %dopar% {
  #library("forecast")
  source('D:/Research/where2test/forecastw2t/R/errorplots.R')
  
  print(i)
  #Extract data for focal region
  DATASET <- subset(X, X$region == i, select=c("date", "cases"))
  
  #Extract dates and case numbers
  DDATES <- as.Date(DATASET$date)
  DCASES <- DATASET$cases
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)
  #Create an empty vector to store optimization results
  RES <- c()
  
  #Grid search over values of the mixing parameter, m, in the combination forecast model
  for(j in seq(0, 1, 0.05)){
    
    #Calculate the raw prediction errors by rolling cross-validation
    #CV starts at week 28, which is at the beginning of the second wave
    CVE <- tsCV(DCASES, combination_daily, h=H, m=j, tests=TESTS, bct=BCT, initial=7*28)
    
    #Error metric to minimize
    #Mean absolute prediction error, MAPE
    ERR <- mape(CVE, DCASES)
    
    #Store model parameters and resulting error value
    RES <- rbind(RES, c(j, sum(ERR)))
    
  }
  
  #Identify and store the parameters associated with the min error
  MINROW <- which.min(RES[,2])
  MINM <- RES[MINROW, 1]
  
  #Create data structure to return to MODS
  c(i, MINM)
}
print(MODS_daily_0)
MODS_daily_0 <- t(data.frame(MODS_daily_0))

# WDATA$comb_j <- NA
# Weekly_pred_comb_0 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
#   library("forecast")
#   
#   #i<-1
#   #Get daily data for focal region
#   DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
#   DDATASET$date <- as.Date(DDATASET$date)
#   DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
#   
#   #Extract dates and case numbers
#   DDATES <- as.Date(head(DDATASET$date, n=-6))
#   DCASES <- head(DDATASET$cases, n=-6)
#   
#   #Cast the case counts as a timeseries with weekly frequency
#   DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
#                frequency = 7)
#   
#   #Probably need to settle on a consistent arima model to use to avoid overfitting
#   #Until then, use auto.arima
#   #ARIMA models behave weirdly when they have too much data to chew on
#   #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
#   #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
#   
#   #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
#   
#   #This is the snaive model formulated as an arima model so that it is compatible with sim code below
#   #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
#   #summary(SARIMA)
#   
#   #Fit the SARIMA model with an automatic box-cox transformation
#   #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
#   #MODEL IS FOUND
#   
#   for(j in seq(0, 1, 0.05)){
#     t <- 0
#     n <- 1
#     DCASES_sub <- head(DCASES, n = 100)
#     while (length(DCASES_sub) == 100) {
#       #SIM <- combination_daily(DCASES_sub, h=HD, m=MODS_daily[i, 2])$mean
#       SIM <- combination_daily(DCASES_sub, h=HD, m=j)$mean
#         
#       NCM <- c(tail(DCASES_sub, n=(7-HDNC)), SIM)
#       WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- colSums(matrix(NCM, nrow = 7))
#       WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'comb_j'] <- j
#       #WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
#       
#       DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
#       
#       t <- t + 4
#       n <- n + 1
#     }
#     if (j==0) temp <- WDATA[WDATA$region == CODES[i],]
#     else temp<- rbind(temp,WDATA[WDATA$region == CODES[i],])
#   }
#   temp
# }
# 
# rm(mape)
# library(ie2misc)
# combine_stats <- data.frame(matrix(NA,nrow=14,ncol=20))
# for (i in 1:14){
#   for(j in 1:20){
#     temp <- Weekly_pred_comb_0[Weekly_pred_comb_0$region == unique(Weekly_pred_comb_0$region)[i] &
#                                  Weekly_pred_comb_0$comb_j == seq(0, 1, 0.05)[j],]
#     combine_stats[i,j] <- mape(temp$pred+1,temp$cases+1,na.rm = TRUE)
#   }
# }
# 
# combine_stats$region <- unique(Weekly_pred_comb_0$region)
# colnames(combine_stats) <- seq(0, 1, 0.05)
# MODS_daily <- as.numeric(colnames(combine_stats)[apply(combine_stats,1,which.min)])

Daily_pred_comb <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  DDATASET$date <- as.Date(DDATASET$date)
  DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
  
  #Extract dates and case numbers
  DDATES <- as.Date(head(DDATASET$date, n=-6))
  DCASES <- head(DDATASET$cases, n=-6)
  
  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)
  
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  t <- 0
  n <- 1
  DCASES_sub <- head(DCASES, n = 100)
  while (length(DCASES_sub) == 100) {
    SIM <- combination_daily(DCASES_sub, h=HD, m=MODS_daily_0[i, 2])$mean
    #SIM <- combination_daily(DCASES_sub, h=HD, m=MODS_daily[i])$mean
    
    NCM <- c(tail(DCASES_sub, n=(7-HDNC)), SIM)
    DDATA_forecast[DDATA_forecast$region == CODES[i],][(t+1):(t+28),'pred'] <- NCM
    #WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'comb_j'] <- j
    #WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
    
    DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
    
    t <- t + 28
    n <- n + 1
  }
  DDATA_forecast[DDATA_forecast$region == CODES[i],]
}

Weekly_pred_comb <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  DDATASET$date <- as.Date(DDATASET$date)
  DDATASET <- DDATASET[DDATASET$date > as.Date('2020-09-14'),]
  
  #Extract dates and case numbers
  DDATES <- as.Date(head(DDATASET$date, n=-6))
  DCASES <- head(DDATASET$cases, n=-6)
  
  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)
  
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  t <- 0
  n <- 1
  DCASES_sub <- head(DCASES, n = 100)
  while (length(DCASES_sub) == 100) {
    SIM <- combination_daily(DCASES_sub, h=HD, m=MODS_daily_0[i, 2])$mean
    #SIM <- combination_daily(DCASES_sub, h=HD, m=MODS_daily[i])$mean
    
    NCM <- c(tail(DCASES_sub, n=(7-HDNC)), SIM)
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- colSums(matrix(NCM, nrow = 7))
    #WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'comb_j'] <- j
    #WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
    
    DCASES_sub <- head(tail(DCASES, n = -28*n), n = 100)
    
    t <- t + 4
    n <- n + 1
  }
  WDATA[WDATA$region == CODES[i],]
}

#Weekly data
WDATA <- read.csv(file=paste0("Data/cases[RKI]_", eva_region, "_weekly.csv"))
WDATA$date <- as.Date(WDATA$date)
WDATA <- WDATA[WDATA$region == eva_region_id,]
WDATA <- WDATA[WDATA$date > "2020-09-06",]
WDATA$pred <- NA

Weekly_pred_arima_weekly_0 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("date", "cases"))
  WDATASET$date <- as.Date(WDATASET$date)
  
  #Extract dates and case numbers
  WDATES <- as.Date(WDATASET$date)
  WCASES <- WDATASET$cases
  
  #Cast the case counts as a timeseries with weekly frequency
  WCASES <- ts(WCASES, start=c(2020, as.numeric(format(WDATES[1], "%j"))),
               frequency = 7)
  
  HD <- 4
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  
  t <- 15
  WCASES_sub <- head(WCASES, n = 15)
  while (length(WCASES_sub) < length(WCASES)-3) {
    ARIMA <- Arima(WCASES_sub, order=c(1, 1, 0),lambda="auto")
    #Generate forecasts via simulation on a daily level with the SARIMA model
    #SIM <- NULL
    #for(j in 1:NREP) {
    #  SIM <- rbind(SIM, simulate(ARIMA, future=TRUE, nsim=HD, level=c(95)))
    #}
    SIM <- forecast(Arima(WCASES_sub, order=c(1, 1, 0)), HD, level=c(95))
    
    NCM <- SIM$mean #colMeans(SIM)
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- NCM
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
    
    t <- t + 4
    WCASES_sub <- head(WCASES, n = t)
  }
  WDATA[WDATA$region == CODES[i],]
}

Weekly_pred_hw_weekly_0 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("date", "cases"))
  WDATASET$date <- as.Date(WDATASET$date)
  
  #Extract dates and case numbers
  WDATES <- as.Date(WDATASET$date)
  WCASES <- WDATASET$cases
  
  #Cast the case counts as a timeseries with weekly frequency
  WCASES <- ts(WCASES, start=c(2020, as.numeric(format(WDATES[1], "%j"))),
               frequency = 7)
  
  HD <- 4
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  
  t <- 15
  WCASES_sub <- head(WCASES, n = 15)
  while (length(WCASES_sub) < length(WCASES)-3) {
    SIM <- holt(WCASES_sub, h=HD, damped=T, level=c(95), lambda="auto")$mean #seasonal="additive", 
    
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- SIM
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
    
    t <- t + 4
    WCASES_sub <- head(WCASES, n = t)
  }
  WDATA[WDATA$region == CODES[i],]
}

library(forecastW2T)
MODS <- mix(WDATA, 4, BCT = T, CORES = nCores[1]-1)
print(MODS)
MODS <- t(data.frame(MODS))

Weekly_pred_comb_weekly_0 <- foreach(i=1:L.CODES, .combine='rbind') %dopar% {
  library("forecast")
  source('D:/Research/where2test/forecastw2t/R/models.R')
  
  #i<-1
  #Get daily data for focal region
  WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("date", "cases"))
  WDATASET$date <- as.Date(WDATASET$date)
  
  #Extract dates and case numbers
  WDATES <- as.Date(WDATASET$date)
  WCASES <- WDATASET$cases
  
  #Cast the case counts as a timeseries with weekly frequency
  WCASES <- ts(WCASES, start=c(2020, as.numeric(format(WDATES[1], "%j"))),
               frequency = 7)
  
  HD <- 4
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  
  
  t <- 15
  WCASES_sub <- head(WCASES, n = 15)
  while (length(WCASES_sub) < length(WCASES)-3) {
    SIM <- combination(WCASES_sub, h=HD, m = MODS[i, 2])$mean
    
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),'pred'] <- SIM
    WDATA[WDATA$region == CODES[i],][(t+1):(t+4),]
    
    t <- t + 4
    WCASES_sub <- head(WCASES, n = t)
  }
  WDATA[WDATA$region == CODES[i],]
}

stopCluster(cl) 

pred_all <- Reduce(function(dtf1, dtf2) merge(dtf1, dtf2, by = c('region','date','cases'), all.x = TRUE),
                   list(Weekly_pred_arima_weekly,Weekly_pred_hw_weekly,Weekly_pred_comb_weekly,
                        Weekly_pred_sarima,Weekly_pred_hw,Weekly_pred_comb))
pred_all <- pred_all[, (names(pred_all) != 'population')]
colnames(pred_all)[4:9] <- c('weekly_arima','weekly_holt','weekly_comb','daily_sarima','daily_hw','daily_comb')
pred_all <- pred_all[pred_all$date>as.Date('2020-12-20'),]

pred_all$day <- rep(1:4,nrow(pred_all)/4+1)[1:nrow(pred_all)]

detach("package:forecastW2T", unload=TRUE)
library(ie2misc)
CODES <- unique(pred_all$region)
L.CODES <- length(CODES)

stats_mape <- data.frame(matrix(NA,nrow=4*L.CODES,ncol=8))
colnames(stats_mape) <- c('day','region',colnames(pred_all)[4:9])
stats_mape$day <- rep(1:4,L.CODES)
stats_mape$region <- rep(CODES,each=4)
for (s in 1:L.CODES){
  for (m in colnames(pred_all)[4:9]){
    for (i in 1:4){
      temp <- pred_all[pred_all$region==CODES[s]&pred_all$day==i,]
      temp <- temp[complete.cases(temp),]
      stats_mape[stats_mape$region==CODES[s]&stats_mape$day==i,m] <- mape(temp[,m]+1,temp$cases+1,na.rm = TRUE)
    }
  }
}

library(ggplot2)
library(reshape2)
a <- ggplot(melt(stats_mape,id.vars = c('day','region'))) +
  geom_boxplot(aes(x=variable,y=value)) +
  facet_wrap(~day,nrow = 1) +
  #coord_cartesian(ylim=c(0,800)) +
  xlab('Methods') +
  ylab('MAPE') +
  theme(legend.position = 'right',legend.title=element_text(size=12),
                      axis.text=element_text(size=15),axis.text.x=element_text(angle=90,vjust = 0.3),
                      axis.title=element_text(size=20,face="bold"), legend.text=element_text(size=12),
                      plot.title = element_text(hjust = 0.5,size=20,face="bold"),
                      strip.text.x = element_text(size=15),
                      strip.background = element_rect(colour=NA, fill=NA))+
  theme(panel.grid.major = element_blank(), panel.grid.minor = element_blank(),
        panel.background = element_blank(), axis.line = element_line(colour = "black"),
        panel.border = element_rect(colour = "black", fill=NA, size=1))

png(file=paste0("D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Figures/boxplot_", eva_region, ".png"),
    width=8, height=4,res=300,units = 'in')
print(a)
dev.off()

a <- ggplot(melt(pred_all[pred_all$region==eva_region_id,],id.vars = c("region","date","day"))) +
  geom_line(aes(x=date,y=value,group=variable,col=variable,linetype=variable),size=0.5)+
  #coord_cartesian(ylim=c(0,80000)) +
  xlab('Date') +
  ylab('Cases') +
  scale_linetype_manual(values=c("solid","solid","solid","solid", "dashed","dashed","dashed"))+
  scale_color_manual(values=c('black','blue','red','cyan','blue','red','cyan'))+
  theme(legend.position = 'right',legend.title=element_text(size=12),
        axis.text=element_text(size=15),#axis.text.x=element_text(angle=90),
        axis.title=element_text(size=20,face="bold"), legend.text=element_text(size=12),
        plot.title = element_text(hjust = 0.5,size=20,face="bold"),
        strip.text.x = element_text(size=15),
        strip.background = element_rect(colour=NA, fill=NA))+
  theme(panel.grid.major = element_blank(), panel.grid.minor = element_blank(),
        panel.background = element_blank(), axis.line = element_line(colour = "black"),
        panel.border = element_rect(colour = "black", fill=NA, size=1))

png(file=paste0("D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Figures/timeseries_", eva_region, ".png"),
    width=6, height=4,res=300,units = 'in')
print(a)
dev.off()

pred_daily <- Reduce(function(dtf1, dtf2) merge(dtf1, dtf2, by = c('region','date','cases','population'), all.x = TRUE),
                   list(Daily_pred_sarima,Daily_pred_hw,Daily_pred_comb))
pred_daily <- pred_daily[, (names(pred_daily) != 'population')]
colnames(pred_daily)[4:6] <- c('daily_sarima','daily_hw','daily_comb')
pred_daily <- pred_daily[pred_daily$date>as.Date('2020-12-20'),]

pred_all$date <- as.POSIXct(pred_all$date)
pred_daily$date <- as.POSIXct(pred_daily$date)
a <- ggplot() +
  geom_line(data=melt(pred_daily[pred_daily$region==eva_region_id,],id.vars = c("region","date")),
            aes(x=date,y=value,group=variable,col=variable),size=0.5)+
  geom_line(data=melt(pred_all[pred_all$region==eva_region_id,c("region","date",'cases','daily_sarima','daily_hw','daily_comb')],id.vars = c("region","date")),
            aes(x=date,y=value,group=variable,col=variable),size=0.5)+
  #coord_cartesian(ylim=c(0,80000)) +
  xlab('Date') +
  ylab('Cases') +
  scale_color_manual(values=c('black','blue','red','cyan'))+
  theme(legend.position = 'right',legend.title=element_text(size=12),
        axis.text=element_text(size=15),#axis.text.x=element_text(angle=90),
        axis.title=element_text(size=20,face="bold"), legend.text=element_text(size=12),
        plot.title = element_text(hjust = 0.5,size=20,face="bold"),
        strip.text.x = element_text(size=15),
        strip.background = element_rect(colour=NA, fill=NA))+
  theme(panel.grid.major = element_blank(), panel.grid.minor = element_blank(),
        panel.background = element_blank(), axis.line = element_line(colour = "black"),
        panel.border = element_rect(colour = "black", fill=NA, size=1))

png(file=paste0("D:/OneDrive/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Figures/timeseries_daily_", eva_region, ".png"),
    width=6, height=4,res=300,units = 'in')
print(a)
dev.off()
