#Need to expand for all regions, parallelize via foreach
#Ideally, this would work entirely from the daily data and
#aggregate to weekly as necessary


library("ggplot2")
library("foreach")
library("doParallel")
library("remotes")
library("forecast")
#library("forecastW2T")

setwd("F:/Research/where2test/com.casus.env.where2test.forecasting/TestSarima/")

#Daily data
DDATA <- read.csv(file="Data/cases[RKI]_saxony_daily.csv")

CODES <- unique(DDATA$region)
L.CODES <- length(CODES)

#Number of obs to use when fitting/selecting ARIMA model
#The last NOBS values of the timeseries are used
#ARIMA MODELS START TO PERFORM POORLY WHEN THEY ARE FIT TO
#TOO MANY OBS. THE IDEA IS TO PREVENT THAT FROM HAPPENING
#BY ONLY USING PART OF THE TIMESERIES, BUT THERE IS NO EXACT
#BEST NOBS.
NOBS <- 100

#Number of sims. This should be large
NREP <- 10000

#This is how many days of data will be ignored at the end of the dataset
HDNC <- 4

#Length of the forecast horizon in weeks, excluding nowcast
HW <- 3

#Overall forecast horizon in days including nowcast
HD <- HW*7 + HDNC

#MODS <- data.frame(matrix(NA,nrow=L.CODES,ncol=2))
#MODS[,1] <- CODES

comp <- data.frame(matrix(NA,nrow=L.CODES,ncol=5))
row.names(comp) <- CODES
colnames(comp) <- c('order','seasonal','lambda','AICc_SARIMA','AICc_auto')
# comp_seasonal_adj <- data.frame(matrix(NA,nrow=L.CODES,ncol=5))
# row.names(comp_seasonal_adj) <- CODES
# colnames(comp_seasonal_adj) <- c('order','seasonal','lambda','AICc_SARIMA','AICc_auto')
#Loop over regions
pdf("F:/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/SARIMA_vs_auto.pdf")
for (i in 1:L.CODES) {
  #i<-1
  #Get daily data for focal region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  
  #Extract dates and case numbers
  DDATES <- as.Date(head(DDATASET$date, n=-6))
  DCASES <- head(DDATASET$cases, n=-6)
  
  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))),
               frequency = 7)
  dt <- tail(DCASES+1, n=100)
  autoplot(dt)
  dt %>% diff() %>% ggtsdisplay(main=paste("Time plot and ACF and PACF plots for the differenced infected cases at",CODES[i]))
  
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  SARIMA <- Arima(dt, order=c(0,1,2),
                  seasonal=c(0,1,1), lambda="auto")
  comp[i,'AICc_SARIMA'] <- SARIMA$aicc
  comp[i,'lambda'] <- SARIMA$lambda[1]
  checkresiduals(SARIMA)
  
  par <- auto.arima(dt)$arma
  comp[i,'order'] <- paste(par[1],par[6],par[2],sep=',')
  comp[i,'seasonal'] <- paste(par[3],par[7],par[4],sep=',')
  
  #BoxCox(dt, lambda = BoxCox.lambda(dt)) %>% diff() %>% ggtsdisplay(main="")
  SARIMA <- Arima(dt, order=c(par[1],par[6],par[2]),
                  seasonal=c(par[3],par[7],par[4]), lambda="auto")
  comp[i,'AICc_auto'] <- SARIMA$aicc
  checkresiduals(SARIMA)
  
  # dt %>% stl(s.window='periodic') %>% seasadj() -> dtadj
  # autoplot(dtadj)
  # dtadj %>% diff() %>% ggtsdisplay(main=paste("Time plot and ACF and PACF plots for the differenced\nseasonally adjusted infected cases at",CODES[i]))
  # 
  # SARIMA <- Arima(dtadj, order=c(0,1,2),
  #                 seasonal=c(0,1,1), lambda="auto")
  # comp_seasonal_adj[i,'AICc_SARIMA'] <- SARIMA$aicc
  # comp_seasonal_adj[i,'lambda'] <- SARIMA$lambda[1]
  # checkresiduals(SARIMA)
  # 
  # par <- auto.arima(dtadj)$arma
  # comp_seasonal_adj[i,'order'] <- paste(par[1],par[6],par[2],sep=',')
  # comp_seasonal_adj[i,'seasonal'] <- paste(par[3],par[7],par[4],sep=',')
  # 
  # #BoxCox(dt, lambda = BoxCox.lambda(dt)) %>% diff() %>% ggtsdisplay(main="")
  # SARIMA <- Arima(dtadj, order=c(par[1],par[6],par[2]),
  #                 seasonal=c(par[3],par[7],par[4]), lambda="auto")
  # comp_seasonal_adj[i,'AICc_auto'] <- SARIMA$aicc
  # checkresiduals(SARIMA)
}  
dev.off()

write.csv(comp,'F:/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/SARIMA_vs_auto.csv')
#write.csv(comp_seasonal_adj,'F:/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/SARIMA_vs_auto_seasonal_adj.csv')