library('RPostgres')
library('dplyr')
library('ggplot2')
library('KernSmooth')
library('sys')
#library("remotes")
#remotes::install_gitlab("calabr29/forecastw2t@develop", host="gitlab.hzdr.de", auth_token="Nbyh89kzgRty5ftGsMxK", force = TRUE)
library("forecastW2T")

args=commandArgs(trailingOnly=TRUE)

# test if there is at least one argument: if not, return an error
if (length(args)==0) {
  stop("At least one argument must be supplied", call.=FALSE)
}

setwd("~/")
w2tforecasting <- paste("Workspaces", "com.casus.env.where2test.forecasting", sep="/")
setwd(w2tforecasting)
getwd()
today <- Sys.Date()
date <- format(today, format="%Y_%m_%d")
folder_name <- paste("Week", date, sep="_")
print(folder_name)
if (!dir.exists(folder_name)) {dir.create(folder_name)}

setwd(folder_name)
#connections
con <- dbConnect(drv = Postgres(),
                 user = "where2test",
                 password = "P4rtyP4rr0t",
                 host = "where2testadm.hzdr.de",
                 port = 32771,
                 dbname="db_where2test"
)

sql_querry <- paste ("select", "*", "from", args[1], ";", sep=" ")
sql_querry_daily <- paste ("select", "*", "from", args[2], ";", sep=" ")

print("csv file created. Forecasting process is running")
#set dataframe using the querry of sql
DATAW <- dbGetQuery(con, sql_querry)
DATAD <- dbGetQuery(con, sql_querry_daily)

if (!dir.exists(args[3])) {dir.create(args[3])}

setwd(args[3])
filename <- paste(substr(args[1], 6, nchar(args[1])), "csv", sep=".")
filename_daily <- paste(substr(args[2], 6, nchar(args[2])), "csv", sep=".")
write.csv(DATAW, filename, row.names = FALSE)
write.csv(DATAD, filename_daily, row.names = FALSE)
dbDisconnect(con)

#Create necessary directories for storing results
if (!dir.exists("ForecastsDaily")) {dir.create("ForecastsDaily")}
setwd("ForecastsDaily")
if (!dir.exists("Error_Plots")) {dir.create("Error_Plots")}
if (!dir.exists("Figures")) {dir.create("Figures")}

#Choose what you want to predict. Again, you must recreate the above-created
#directories for each set of predictions
#Set TARGET <- 1 for Saxony cases
#Set TARGET <- 2 for Czechia cases
#Set TARGET <- 3 for Czechia tests
#Set TARGET <- 4 for Saxony vaccine
TARGET <- as.integer(args[4])
print(paste("Target No", TARGET, sep=" "))
home <- paste("/home", "wildan", sep="/")
INPATHDaily = paste(home, w2tforecasting, folder_name,args[3], filename_daily, sep="/")
INPATHWeekly = paste(home, w2tforecasting, folder_name,args[3], filename, sep="/")
print(INPATHDaily)
print(INPATHWeekly)
DATAREADD <- read.csv(file=INPATHDaily)
DATAREADW <- read.csv(file=INPATHWeekly)

zero_days <- function(df){
  tmp <- df[df$region==unique(df$region)[1],'cases']
    d <- 1
    while (sum(tail(tmp,n=d))==0){
	        d = d + 1
      }
      return(d)
}

FILTERDATAD <- subset(DATAREADD, as.Date(date) < Sys.Date() - zero_days(DATAREADD)  &  as.Date(date) >= Sys.Date() - 100)
#Get the mix parameter estimates for the current data file
#Set number of processor cores (CORES) appropriately for your machine
MODS <- mix_daily(FILTERDATAD, H=28, CORES=8)
MODSW <- mix(DATAREADW, H=4, CORES=8)

#Get the forecasts for the current data file, conditional on appropriate mix param estimates
#Set number of processor cores (CORES) appropriately for your machine
FORECASTS <- forecastW2T_daily(DATAREADD, DATAREADW, H=4, MODS_daily=MODS, CORES=8, MODS=MODSW)
OUTPATH <- paste(home, w2tforecasting, folder_name,args[3], sep="/")
switch(TARGET,
       write.csv(FORECASTS, file=paste0(OUTPATH, "/Saxony_Cases_Forecasts_Daily_", toString(Sys.Date()), ".csv"), row.names = F),
       write.csv(FORECASTS, file=paste0(OUTPATH, "/Czechia_Cases_Forecasts_Daily_", toString(Sys.Date()), ".csv"), row.names = F),
       write.csv(FORECASTS, file=paste0(OUTPATH, "/Czechia_Tests_Forecasts_Daily_", toString(Sys.Date()), ".csv"), row.names = F),
       write.csv(FORECASTS, file=paste0(OUTPATH, "/Saxony_Vaccination_Forecasts_Daily_", toString(Sys.Date()), ".csv"), row.names = F),
       write.csv(FORECASTS, file=paste0(OUTPATH, "/Bundesland_Hospital_Forecasts_Daily_", toString(Sys.Date()), ".csv"), row.names = F),
       write.csv(FORECASTS, file=paste0(OUTPATH, "/Dolnoslaski_Cases_Forecasts_Daily_", toString(Sys.Date()), ".csv"), row.names = F)
)

