
library("remotes")
library("forecastW2T")
library("forecast")
library("ggplot2")

#Calculate mean average percent error from raw errors and data
#mape <- function(err, data) {100*colMeans(abs(err/data), na.rm = T)}
mape <- function(err, data){
  
  val <- abs(err/data)
  val[!is.finite(val)] <- NA
  val <- 100*colMeans(val, na.rm = T)
  
  return(val)
}


#Read in a data file. Do this only one at a time
DATA <- read.csv(file="data_training_hospital.csv")

#Check if the data are on cases or tests. If on tests, note that and then change
#the column title to cases, which is what the following code assumes
if(names(DATA)[4]=="hosp"){
  TESTS <- TRUE
  names(DATA)[4] <- "hosp_cases"
} else{
  TESTS <- FALSE
}

#Extract the region codes from the data
CODES <- unique(DATA$region)
L.CODES <- 1:length(CODES)


#Loop over regions, performing a forecast for each region
DATASET <- subset(DATA, DATA$region == CODES[1], select=c("date", "hosp_cases", "cases" ))
#Extract dates and case numbers
DATES <- as.Date(DATASET$date)
CASES <- DATASET$cases


# If we pass external_regressor (in our case 'saxony incidences') as a covariate into the forecast, 
# h will be disregarded and we will get a forecast for length(external_regressor)
# To overcome this and forecast four weeks in advance, we must provide forecasted 
# external regressor data (in our case 'saxony incidences')
# for the upcoming four weeks, so that length(new_regressor) == 4 as the covariate of the hospitalization
# forecast.

# Fit the auto.arima/Arima/Holts/Combination model on the saxony incidences and generate the forecast for saxony incidences of the next four weeks

#Length of the forecast horizon in weeks
HW <- 4

#cases_fit_autoarima <- auto.arima(DATASET$cases, lambda="auto")
cases_fit_arima      <- Arima(DATASET$cases, order=c(1,1,0), lambda="auto")
FORECAST_cases_arima <- forecast(cases_fit_arima, h=HW, level=c(95))


# Fit the ARIMA model on the saxony hospitalization incidences 
hosp_cases_fit <- Arima(DATASET$hosp_cases, xreg = DATASET$cases, order = c(1,1,0), lambda="auto")

#generate the forecast for saxony hospitalization incidences of the next four weeks using the covariate as the mean forecast of the saxony incidences
FORECAST_hosp_cases <- forecast(hosp_cases_fit, xreg=FORECAST_cases_arima$mean, level=c(95))


#Title for graphs with model parameter info
#The if logic below automatically changes the fig title based on whether the data
#are on cases or tests
TITLE <- paste0("region= 14000", ", ", "Arima model: (1,1,0)", ", ", "mix= None")

#Plot of data and forecast over time
P1 <- autoplot(FORECAST_hosp_cases) + ggtitle(TITLE)
P1
#save the plot to file
#Write plot to file
ggsave(plot = P1,
       filename = file.path("Forecast_Plots", paste0("Forecast_meancov", ".png")),
       device = "png")

