#Need to expand for all regions, parallelize via foreach
#Ideally, this would work entirely from the daily data and 
#aggregate to weekly as necessary


library("ggplot2")
library("foreach")
library("doParallel")
library("remotes")
library("forecast")
library("forecastW2T")

#Register a cluster for parallelized loop
nCores <- detectCores()
cl<- makeCluster(nCores[1]-1) #not to overload your computer
registerDoParallel(cl)

setwd("C:/Workspaces/com.casus.env.where2test.forecasting/TestSarima/")

#Weekly data
WDATA <- read.csv(file="Data/cases[RKI]_saxony_weekly.csv")

#Daily data
DDATA <- read.csv(file="Data/cases[RKI]_saxony_daily.csv")

CODES <- unique(WDATA$region)
L.CODES <- length(CODES)

#Number of obs to use when fitting/selecting ARIMA model
#The last NOBS values of the timeseries are used
#ARIMA MODELS START TO PERFORM POORLY WHEN THEY ARE FIT TO
#TOO MANY OBS. THE IDEA IS TO PREVENT THAT FROM HAPPENING
#BY ONLY USING PART OF THE TIMESERIES, BUT THERE IS NO EXACT
#BEST NOBS.
NOBS <- 100

#Number of sims. This should be large
NREP <- 10000

#This is how many days of data will be ignored at the end of the dataset
HDNC <- 4

#Length of the forecast horizon in weeks, excluding nowcast
HW <- 3

#Overall forecast horizon in days including nowcast
HD <- HW*7 + HDNC

#Loop over regions
foreach(i=1:L.CODES) %dopar% {  
  library("forecast")
  library("forecastW2T")

  #i<-1 
  #Get daily data for focal region
  DDATASET <- subset(DDATA, DDATA$region == CODES[i], select=c("date", "cases"))
  
  #Extract dates and case numbers
  DDATES <- as.Date(head(DDATASET$date, n=-6))
  DCASES <- head(DDATASET$cases, n=-6)
  
  #Cast the case counts as a timeseries with weekly frequency
  DCASES <- ts(DCASES, start=c(2020, as.numeric(format(DDATES[1], "%j"))), 
               frequency = 7)
  
  #Probably need to settle on a consistent arima model to use to avoid overfitting
  #Until then, use auto.arima
  #ARIMA models behave weirdly when they have too much data to chew on
  #As a pragmatic workaround, take the last 100 days for fitting/selecting the model
  #SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)
  
  #ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)
  
  #This is the snaive model formulated as an arima model so that it is compatible with sim code below
  #ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
  #summary(SARIMA)
  
  #Fit the SARIMA model with an automatic box-cox transformation
  #THIS WAS JUST ONE PARTICULAR SARIMA MODEL AND COULD BE CHANGED IF A BETTER
  #MODEL IS FOUND
  SARIMA <- Arima(tail(DCASES+1, n=100), order=c(0,1,2), 
                  seasonal=c(0,1,1), lambda="auto")
  
  #Generate forecasts via simulation on a daily level with the SARIMA model
  SIM <- NULL
  for(j in 1:NREP) {
    SIM <- rbind(SIM, simulate(SARIMA, future=TRUE, nsim=HD))
  }
  
  #Quick check if average (within time) prediction results look reasonable
  #plot(colMeans(SIM), type="l")
  
  #Extract the first 3 days of the final week in the data
  PREWEEK <- sum(tail(DCASES, n=(7-HDNC)))
  
  #Extract the nowcast part of the predictions
  NC <- rowSums(SIM[, 1:HDNC])
  
  #Create the nowcast by combining the last 4 days of data and the first 3 days of predictions
  NOWCAST <- PREWEEK + NC
  
  #Get the mean and PI limits for the nowcast
  NCM <- mean(NOWCAST)
  NCL <- quantile(NOWCAST, 0.025)
  NCU <- quantile(NOWCAST, 0.975)
  
  #Aggregate the daily predictions for week 1 after the nowcast
  W1 <- rowSums(SIM[, 5:11])
  
  #Get the mean and PI limits for the week 1 forecast
  W1M <- mean(W1)
  W1L <- quantile(W1, 0.025)
  W1U <- quantile(W1, 0.975)
  
  #Aggregate the daily predictions for week 2 after the nowcast
  W2 <- rowSums(SIM[, 12:18])
  
  #Get the mean and PI limits for the week 2 forecast
  W2M <- mean(W2)
  W2L <- quantile(W2, 0.025)
  W2U <- quantile(W2, 0.975)
  
  #Aggregate the daily predictions for week 3 after the nowcast
  W3 <- rowSums(SIM[, 19:25])
  
  #Get the mean and PI limits for the week 3 forecast
  W3M <- mean(W3)
  W3L <- quantile(W3, 0.025)
  W3U <- quantile(W3, 0.975)
  
  #Prediction means and lower and upper interval limits
  PDM <- c(NCM, W1M, W2M, W3M)
  PDL <- c(NCL, W1L, W2L, W3L) 
  PDU <- c(NCU, W1U, W2U, W3U)
  
  #Weekly data stuff starts here
  WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("date", "cases"))

  #THIS IS TEMPORARY. REMOVE ONCE DATA ARE IN CONSISTENT FORMAT
  WDATASET  <- WDATASET[-nrow(WDATASET), ]

  WCASES <- WDATASET$cases
  WDATES <- as.Date(WDATASET$date)

  #Get the end date of the data
  ED <- WDATES[length(WDATES)]

  #Generate a vector of forecast dates
  #DATES.PRD <- c(DATES[length(DATES)], seq(ED + 7, ED + 7*HW, 7))
  #DATES.PRD <- seq(ED + 7, ED + 7*(HW+1), 7)
  #DATES.PRD
  #Generate a vector of forecast dates
  #FORE.DATES <- seq(ED + 7, ED + 7*H, 7)
  DATES.PRD <- c(ED + 7, ED + 14, ED + 21, ED + 28)

  #Create output dataframes for data and forecasts
  L.WDATA <- length(WCASES)

  #Fill in prediction interval columns with NAs to keep same structure as forecast output
  DATA.OUT <- data.frame(
    region = rep(CODES[i], L.WDATA),
    date = WDATES,
    m = WCASES,
    l95 = rep(NA, L.WDATA),
    u95 = rep(NA, L.WDATA)
  )

  FORE.OUT <- data.frame(
    region = rep(CODES[i], HW+1),
    date = DATES.PRD,
    m = PDM,
    l95 = PDL,
    u95 = PDU
  )

  #Collate the data and forecast outputs
  #FORECASTS <- rbind(FORECASTS, DATA.OUT, FORE.OUT)

  if (!dir.exists("Predictions")) {dir.create("Predictions")}
  if (!dir.exists("Figures")) {dir.create("Figures")}
  write.csv(rbind(DATA.OUT, FORE.OUT), file=paste0("Predictions/", toString(CODES[i])), row.names = F)

  #Need to output figure via graphics device
  png(file=paste0("Figures/", toString(CODES[i]), ".png"),
     width=1200, height=700)
  
  par(mar=c(5,5,2,0))
  
  #Plot weekly data
  plot(WDATES, WCASES, xaxt='n', type="l", ylab="Cases",
       xlab="Date", ylim=c(0, max(WCASES, W3U)),
       xlim=as.Date(c(WDATES[1], tail(DATES.PRD, n=1))),
       cex.lab=1.5, lwd=2, main=toString(CODES[i]))
  
  #Add the lines for the mean predictions
  lines(DATES.PRD, PDM, col="blue", lwd=2)

  #Add the PIs as a blue shaded region
  polygon(x = c(DATES.PRD, rev(DATES.PRD)),
          y = c(PDU, rev(PDL)),
          col =  adjustcolor("dodgerblue", alpha.f = 0.30), border = NA)

  #Add a red point to indicate the nowcast
  points(DATES.PRD[1], NCM, col="red", lwd=2, pch=20)

  #Add the correct dates to the x-axis
  ALL.DATES <- c(WDATES, DATES.PRD)
  axis(1, ALL.DATES, format(ALL.DATES, "%m/%y"), cex.axis = 1)

  dev.off()
}
stopCluster(cl)