#Need to expand for all regions, parallelize via foreach
#Ideally, this would work entirely from the daily data and
#aggregate to weekly as necessary
library("remotes")

remotes::install_gitlab("calabr29/forecastw2t", host="gitlab.hzdr.de", auth_token="nZzyUY6kp-AD51QSAqew")

library("forecastW2T")

library("ggplot2")
library("foreach")
library("doParallel")
library("remotes")
library("forecast")
#library("forecastW2T")

#Register a cluster for parallelized loop
nCores <- detectCores()
cl<- makeCluster(nCores[1]-1) #not to overload your computer
registerDoParallel(cl)

setwd("C:/Workspaces/com.casus.env.where2test.forecasting/TestSarima/")

#Weekly data
WDATA <- read.csv(file="Data/cases[RKI]_saxony_weekly.csv")

#Daily data
DDATA <- read.csv(file="Data/cases[RKI]_saxony_daily.csv")

CODES <- unique(WDATA$region)
L.CODES <- length(CODES)

#Number of obs to use when fitting/selecting ARIMA model
#The last NOBS values of the timeseries are used
#ARIMA MODELS START TO PERFORM POORLY WHEN THEY ARE FIT TO
#TOO MANY OBS. THE IDEA IS TO PREVENT THAT FROM HAPPENING
#BY ONLY USING PART OF THE TIMESERIES, BUT THERE IS NO EXACT
#BEST NOBS.
NOBS <- 100

#Number of sims. This should be large
NREP <- 10000

#This is how many days of data will be ignored at the end of the dataset
HDNC <- 4

#Length of the forecast horizon in weeks, excluding nowcast
HW <- 3

#Overall forecast horizon in days including nowcast
HD <- HW*7 + HDNC

#MODS <- data.frame(matrix(NA,nrow=L.CODES,ncol=2))
#MODS[,1] <- CODES

DDATASET <- subset(DDATA, select=c("region", "date", "cases"))
DDATASET$date <- as.Date(DDATASET$date)
DDATASET <- DDATASET[DDATASET$date > as.Date('2021-11-02'),]

#Extract dates and case numbers
DDATES <- as.Date(head(DDATASET$date, n=-6))
DCASES <- head(DDATASET$cases, n=-6)

#Cast the case counts as a timeseries with weekly frequency
DCASES <- ts(DCASES, start=c(2021, as.numeric(format(DDATES[1], "%j"))),
             frequency = 7)

#Probably need to settle on a consistent arima model to use to avoid overfitting
#Until then, use auto.arima
#ARIMA models behave weirdly when they have too much data to chew on
#As a pragmatic workaround, take the last 100 days for fitting/selecting the model
#SARIMA <- auto.arima(tail(DCASES+1, n=100), approximation=F, lambda="auto", num.cores = nCores)

#ARIMA <- ets(window(DCASES+1, start=2080), model="AAA", lambda="auto", damped=T)

#This is the snaive model formulated as an arima model so that it is compatible with sim code below
#ARIMA <- Arima(CASES+1, order=c(1,1,0), seasonal=c(1,1,2), lambda="auto", biasadj = T)
#summary(SARIMA)

MODS <- mix(DDATASET, H=(HW + 1) * 7, CORES=cl)
print(MODS)

auto_sarima_par <- read.csv('Parameters_Test/SARIMA_vs_auto.csv')
foreach(i=1:L.CODES) %dopar% {
  library("ggplot2")
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  DDATASET <- subset(DDATASET, DDATASET$region == CODES[i], select=c("date", "cases"))
  
  #Calculate raw forecast errors for different methods via rolling CV
  H = (HW + 1) * 7
  source('C:/Workspaces/forecastw2t/R/models.R')
  ERR.NA <- tsCV(DDATASET$cases, fnaive, h = H, initial=1)
  ERR.HT <- tsCV(DDATASET$cases, fholt, h = H, initial=1)
  ERR.AA <- tsCV(DDATASET$cases, farima, h = H, initial=1)
  ERR.A1 <- tsCV(DDATASET$cases, farima110, h = H, initial=1)
  ERR.CB <- tsCV(DDATASET$cases, combination, h = H, m=MODS[i, 2], tests=FALSE, bct=TRUE, initial=1)

  farima <- function(x, h){
    forecast(Arima(x+1, order=c(0, 1, 2), seasonal=c(0,1,1), lambda="auto"), h, level=c(95))
  }
  ERR.SA <- tsCV(DDATASET$cases, farima, h = H, initial=1)
  
  farima_auto <- function(x, h){
    forecast(Arima(x+1, order=c(as.numeric(strsplit(auto_sarima_par$order[i], ',')[[1]])), 
                   seasonal=c(as.numeric(strsplit(auto_sarima_par$seasonal[i], ',')[[1]])), 
                   lambda=auto_sarima_par$lambda[i]), h, level=c(95))
  }

  ERR.SA_auto <- tsCV(DDATASET$cases, farima_auto, h = H, initial=1)
  #Calculate MAPE for different methods
  source('C:/Workspaces/forecastw2t/R/errorplots.R')
  MAPE.NA <- mape(ERR.NA, DDATASET$cases)
  MAPE.HT <- mape(ERR.HT, DDATASET$cases)
  MAPE.AA <- mape(ERR.AA, DDATASET$cases)
  MAPE.A1 <- mape(ERR.A1, DDATASET$cases)
  MAPE.CB <- mape(ERR.CB, DDATASET$cases)
  MAPE.SA <- mape(ERR.SA, DDATASET$cases)
  MAPE.SA_auto <- mape(ERR.SA_auto, DDATASET$cases)
  
  #Create data frame for plotting MAPE by prediction step
  DF <- data.frame(h = 1:H,
                   err.na = MAPE.NA,
                   err.ht = MAPE.HT,
                   err.aa = MAPE.AA,
                   err.a1 = MAPE.A1,
                   err.cb = MAPE.CB,
                   err.sa = MAPE.SA,
                   err.sa_auto = MAPE.SA_auto
  )
  
  #Plot MAPE by prediction step
  TITLE <- paste0("region=", CODES[i])
  P2 <- ggplot(DF, aes(x = 1:H, y = value, color = variable)) +
    
    geom_point(aes(y = err.na, col = "y1"), size=2) +
    geom_point(aes(y = err.ht, col = "y2"), size=2) +
    geom_point(aes(y = err.aa, col = "y3"), size=2) +
    geom_point(aes(y = err.a1, col = "y4"), size=2) +
    geom_point(aes(y = err.cb, col = "y5"), size=2) +
    geom_point(aes(y = err.sa, col = "y6"), size=2) +
    geom_point(aes(y = err.sa_auto, col = "y7"), size=2) +
    
    geom_line(aes(y = err.na, col = "y1")) +
    geom_line(aes(y = err.ht, col = "y2")) +
    geom_line(aes(y = err.aa, col = "y3")) +
    geom_line(aes(y = err.a1, col = "y4")) +
    geom_line(aes(y = err.cb, col = "y5")) +
    geom_line(aes(y = err.sa, col = "y6")) +
    geom_line(aes(y = err.sa_auto, col = "y7")) +
    
    labs(x = "Forecast horizon (days)",
         y = "Mean absolute percentage error") +
    
    theme(axis.title=element_text(size=14)) +
    
    scale_color_discrete(
      name = "Method:",
      labels = c("Naive", "Holt", "AutoArima", "Arima110", "Combination", "SARIMA")
    ) +
    ggtitle(TITLE) +
    theme(plot.title = element_text(lineheight = 0.8, size = 16))
  #Create necessary directories for storing results
  if (!dir.exists("Figures/Error_Plots")) {dir.create("Figures/Error_Plots")}
  #Write error plots to disk
  ggsave(plot = P2,
         filename = file.path("C:/Workspaces/com.casus.env.where2test.forecasting/TestSarima/Figures/Error_Plots", paste0("Error_Daily", CODES[i], ".png")),
         device = "png")
}
stopCluster(cl)

#Weekly error
registerDoParallel(cl)
# #Get the mix parameter estimates for the current data file
# #Set number of processor cores (CORES) appropriately for your machine
# MODS <- mix(DATA, H=4, CORES=nCores[1]-1)
# 
# #Get the forecasts for the current data file, conditional on appropriate mix param estimates
# #Set number of processor cores (CORES) appropriately for your machine
# FORECASTS <- forecastW2T(DATA, H=4, MODS, CORES=nCores[1]-1)
# 
# #Write the forecasts to file, choosing the output names as appropriate
# write.csv(FORECASTS, file=paste0("F:/Research/where2test/com.casus.env.where2test.forecasting/Daily_test_Kai/Figures/Error_Plots/Saxony_Cases_Forecasts_", toString(Sys.Date()), ".csv"), row.names = F)
# stopCluster(cl)
MODS <- mix(WDATA, 4, BCT = T, CORES = nCores[1]-1)
print(MODS)

DDATASET <- subset(DDATA, select=c("region", "date", "cases"))
DDATASET$date <- as.Date(DDATASET$date)
DDATASET <- DDATASET[DDATASET$date > as.Date('2021-11-02'),]

#Extract dates and case numbers
DDATES <- as.Date(head(DDATASET$date, n=-6))
DCASES <- head(DDATASET$cases, n=-6)

#Cast the case counts as a timeseries with weekly frequency
DCASES <- ts(DCASES, start=c(2021, as.numeric(format(DDATES[1], "%j"))),
             frequency = 7)

foreach(i=1:L.CODES) %dopar% {
  library("ggplot2")
  library("forecast")
  
  #i<-1
  #Get daily data for focal region
  WDATASET <- subset(WDATA, WDATA$region == CODES[i], select=c("date", "cases"))
  DDATASET <- subset(DDATASET, DDATASET$region == CODES[i], select=c("date", "cases"))
  
  #Calculate raw forecast errors for different methods via rolling CV
  H = 4
  source('C:/Workspaces/forecastw2t/R/models.R')
  ERR.NA <- tsCV(WDATASET$cases, fnaive, h = H, initial=28)
  ERR.HT <- tsCV(WDATASET$cases, fholt, h = H, initial=28)
  ERR.AA <- tsCV(WDATASET$cases, farima, h = H, initial=28)
  ERR.A1 <- tsCV(WDATASET$cases, farima110, h = H, initial=28)
  ERR.CB <- tsCV(WDATASET$cases, combination, h = H, m=MODS[i, 2], tests=FALSE, bct=TRUE, initial=28)
  H = (HW + 1) * 7
  farima <- function(x, h){
    forecast(Arima(x+1, order=c(0, 1, 2), seasonal=c(0,1,1), lambda="auto"), h, level=c(95))
  }
  ERR.SA <- tsCV(DDATASET$cases, farima, h = H, initial=1)
  farima_auto <- function(x, h){
    forecast(Arima(x+1, order=c(as.numeric(strsplit(auto_sarima_par$order[i], ',')[[1]])), 
                   seasonal=c(as.numeric(strsplit(auto_sarima_par$seasonal[i], ',')[[1]])), 
                   lambda=auto_sarima_par$lambda[i]), h, level=c(95))
  }
  ERR.SA_auto <- tsCV(DDATASET$cases, farima_auto, h = H, initial=1)
  
  #Calculate MAPE for different methods
  source('C:/Workspaces/forecastw2t/R/errorplots.R')
  MAPE.NA <- mape(ERR.NA, WDATASET$cases)
  MAPE.HT <- mape(ERR.HT, WDATASET$cases)
  MAPE.AA <- mape(ERR.AA, WDATASET$cases)
  MAPE.A1 <- mape(ERR.A1, WDATASET$cases)
  MAPE.CB <- mape(ERR.CB, WDATASET$cases)
  MAPE.SA <- mape(ERR.SA, DDATASET$cases)
  MAPE.SA_auto <- mape(ERR.SA_auto, DDATASET$cases)
  
  #Create data frame for plotting MAPE by prediction step
  DF <- data.frame(h = seq(1,H,7),
                   err.na = MAPE.NA,
                   err.ht = MAPE.HT,
                   err.aa = MAPE.AA,
                   err.a1 = MAPE.A1,
                   err.cb = MAPE.CB)
  
  DF_day <- data.frame(h = 1:28,
                   err.sa = MAPE.SA,
                   err.sa_auto = MAPE.SA_auto
  )
  
  DF <- merge(DF,DF_day,by='h',all = TRUE)
  
  #Plot MAPE by prediction step
  TITLE <- paste0("region=", CODES[i])
  P2 <- ggplot(DF, aes(x = 1:H, y = value, color = variable)) +
    
    geom_point(aes(y = err.na, col = "y1"), size=2) +
    geom_point(aes(y = err.ht, col = "y2"), size=2) +
    geom_point(aes(y = err.aa, col = "y3"), size=2) +
    geom_point(aes(y = err.a1, col = "y4"), size=2) +
    geom_point(aes(y = err.cb, col = "y5"), size=2) +
    geom_point(aes(y = err.sa, col = "y6"), size=2) +
    geom_point(aes(y = err.sa_auto, col = "y7"), size=2) +
    
    geom_line(data=DF[!is.na(DF$err.na),], aes(x = h, y = err.na, col = "y1")) +
    geom_line(data=DF[!is.na(DF$err.ht),], aes(x = h, y = err.ht, col = "y2")) +
    geom_line(data=DF[!is.na(DF$err.aa),], aes(x = h, y = err.aa, col = "y3")) +
    geom_line(data=DF[!is.na(DF$err.a1),], aes(x = h, y = err.a1, col = "y4")) +
    geom_line(data=DF[!is.na(DF$err.cb),], aes(x = h, y = err.cb, col = "y5")) +
    geom_line(aes(y = err.sa, col = "y6")) +
    geom_line(aes(y = err.sa_auto, col = "y7")) +
    
    labs(x = "Forecast horizon (days)",
         y = "Mean absolute percentage error") +
    
    theme(axis.title=element_text(size=14)) +
    
    scale_color_discrete(
      name = "Method:",
      labels = c("Naive", "Holt", "AutoArima", "Arima110", "Combination", "SARIMA", "SARIMA_auto")
    ) +
    ggtitle(TITLE) +
    theme(plot.title = element_text(lineheight = 0.8, size = 16))
  
  #Write error plots to disk
  ggsave(plot = P2,
         filename = file.path("C:/Workspaces/com.casus.env.where2test.forecasting/TestSarima/Figures/Error_Plots", paste0("Error_Weekly", CODES[i], ".png")),
         device = "png")
}
stopCluster(cl)
